-------------------------------------------------------------------------------
--! @file       CipherCore_TP.vhd
--! @brief      CipherCore_TP for HS1-SIV
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;
use work.AEAD_pkg.all;
use work.HS1_SIV_pkg.all;

entity CipherCore_TP is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_FW            : integer := 128;   --! Two-Pass FIFO
        G_DBLK_SIZE     : integer := 128;   --! Data
        G_KEY_SIZE      : integer := 128;   --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 4;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := TWO_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic;
        --! Second-pass FIFO
        fdi_data        : in  std_logic_vector(G_FW             -1 downto 0);
        fdi_valid       : in  std_logic;
        fdi_ready       : out std_logic;
        fdo_data        : out std_logic_vector(G_FW             -1 downto 0);
        fdo_valid       : out std_logic;
        fdo_ready       : in  std_logic
    );
end entity CipherCore_TP;

architecture structure of CipherCore_TP is
    signal len_a        : std_logic_vector(G_MAX_LEN            -1 downto 0);
    signal len_d        : std_logic_vector(G_MAX_LEN            -1 downto 0);
    signal npub         : std_logic_vector(96                   -1 downto 0);
    signal tag          : std_logic_vector(128                  -1 downto 0);
    signal to_accum1    : std_logic_vector(128                  -1 downto 0);
    signal from_accum1  : std_logic_vector(128                  -1 downto 0);
    signal from_salsa   : std_logic_vector(128                  -1 downto 0);
    signal to_accum2    : std_logic_vector(64                   -1 downto 0);
    signal to_sum       : std_logic_vector(60                   -1 downto 0);
    signal hash         : std_logic_vector(256                  -1 downto 0);
    
    signal accum_bdi_ready      : std_logic;
    signal accum_bdo_valid      : std_logic;
    signal accum_salsa_ready    : std_logic;
    signal sched_bdi_ready      : std_logic;
    signal sched_bdo_valid      : std_logic;
    signal sched_salsa_ready    : std_logic;    
    signal hash_accum1_end      : std_logic;
    signal hash_accum1_valid    : std_logic;
    signal hash_accum1_ready    : std_logic;
    signal hash_accum2_end      : std_logic;
    signal hash_accum2_valid    : std_logic;
    signal hash_accum2_ready    : std_logic;
    signal wr_nhkey             : std_logic;
    signal is_decrypt           : std_logic;
    signal is_second            : std_logic;
    signal start_hash           : std_logic;
    signal start_output         : std_logic;    
    signal start_salsa          : std_logic;        
    signal salsa_init0          : std_logic;        
    signal init_main_key        : std_logic;
    signal init_ccha_key        : std_logic;
    signal init_nhkey           : std_logic;
    signal init_poly_key        : std_logic;
    signal sel_bus_master       : std_logic;
    signal sel_salsa_master     : std_logic;
    signal sel_fifo             : std_logic;
    signal en_len               : std_logic;
    signal en_npub              : std_logic;
    signal sel_tag              : std_logic;
    signal en_tag_sched         : std_logic;
    signal en_tag_accum1        : std_logic;
    signal salsa_busy           : std_logic;
    signal salsa_valid          : std_logic;
    signal salsa_ready          : std_logic;
    signal output_done          : std_logic;
    signal hash_done            : std_logic;
begin

    process(clk)
    begin
        if rising_edge(clk) then
            if en_len = '1' then
                len_a <= bdi(96+G_MAX_LEN-1 downto 96);
                len_d <= bdi(64+G_MAX_LEN-1 downto 64);
            end if;
            if en_npub = '1' then
                npub  <= bdi(127 downto 32);
            end if;
            if (en_tag_sched = '1' or en_tag_accum1 = '1') then
                if (sel_tag = '1') then
                    tag <= switch_endian_byte(from_salsa, 32);
                else
                    tag <= bdi;
                end if;
            end if;
        end if;
    end process;
    
    msg_auth_valid <= '1' when tag = switch_endian_byte(from_salsa, 32) else '0';
    bdi_ready <= accum_bdi_ready when sel_bus_master = '1' else sched_bdi_ready;
    bdo_valid <= accum_bdo_valid when sel_bus_master = '1' else sched_bdo_valid;
    --! Note: Switch condition for bus mastering here. This is not a typo ...
    salsa_ready <= sched_salsa_ready when sel_salsa_master = '1' else accum_salsa_ready;
    to_accum1 <= fdi_data when (sel_fifo = '1') else bdi;
    fdo_data    <= bdi;
    bdo <= tag when (sel_tag = '1') else from_accum1;

    u_hash_accum1: entity work.HS1_Hash_Accum1
    generic map  (
        G_MAX_LEN   => G_MAX_LEN
    )                   
    port map (          
        clk             => clk                      , 
        rst             => rst                      , 
        --! Input/Output               
        bdi             => to_accum1                , 
        bdo             => from_accum1              , 
        from_salsa      => from_salsa               , 
        data_o          => to_accum2                , 
        --! Ext control                
        salsa_valid     => salsa_valid              , 
        salsa_ready     => accum_salsa_ready        , 
        bdi_valid       => bdi_valid                , 
        bdi_ready       => accum_bdi_ready          , 
        bdo_valid       => accum_bdo_valid          , 
        bdo_ready       => bdo_ready                , 
        fdi_valid       => fdi_valid                ,
        fdi_ready       => fdi_ready                ,
        fdo_valid       => fdo_valid                , 
        --! Control                    
        wr_key          => wr_nhkey                 ,         
        start           => start_hash               , 
        start_output    => start_output             , 
        output_done     => output_done              , 
        init_key        => init_nhkey               , 
        is_decrypt      => is_decrypt               , 
        is_second       => is_second                , 
        len_a           => len_a                    , 
        len_d           => len_d                    , 
        en_tag          => en_tag_accum1            , 
        oend            => hash_accum1_end          , 
        ovalid          => hash_accum1_valid        , 
        oready          => hash_accum1_ready      
    );
    
    u_hash_accum2: entity work.HS1_Hash_Accum2
    port map (
        clk             => clk                      , 
        rst             => rst                      , 
        --! Input/Output
        data_i          => to_accum2(59 downto 0)   , 
        data_o          => to_sum                   , 
        --! Control    
        start           => start_hash               , 
        iend            => hash_accum1_end          , 
        ivalid          => hash_accum1_valid        , 
        iready          => hash_accum1_ready        , 
        oend            => hash_accum2_end          ,
        ovalid          => hash_accum2_valid        ,
        oready          => hash_accum2_ready 
    );
    
    u_hash_sum: entity work.HS1_Hash_Sum
    port map (                         
        clk             => clk                      ,  
        rst             => rst                      ,  
        --! Input/Output               
        key             => from_salsa               ,  
        data_i          => to_sum                   ,  
        hash            => hash                     ,  
        --! Control                    
        start           => start_hash               ,  
        init_pkey       => init_poly_key            ,  
        init_ckey       => init_ccha_key            ,  
        iend            => hash_accum2_end          ,  
        ivalid          => hash_accum2_valid        ,  
        iready          => hash_accum2_ready        ,  
        done            => hash_done          
    );    

    u_salsa: entity work.HS1_SIV_Salsa20
    generic map  (
        G_MAX_LEN   => G_MAX_LEN   
    )       
    port map (
        clk             => clk                      ,   
        rst             => rst                      ,   
        --! Input/out_r                
        key             => key                      ,   
        npub            => npub                     ,   
        hash            => hash                     ,   
        yy              => from_salsa               ,   
        --! Control                    
        start           => start_salsa              ,   
        init0           => salsa_init0              ,   
        init_key        => init_main_key            ,   
        busy            => salsa_busy               ,   
        ovalid          => salsa_valid              ,   
        oready          => salsa_ready      
    );

    u_sched: entity work.HS1_SIV_Sched
    port map   (                                   
        clk              => clk                     ,
        rst              => rst                     ,
        --! Ext. Control                           
        bdi_valid        => bdi_valid               ,
        bdi_ready        => sched_bdi_ready         ,
        bdo_valid        => sched_bdo_valid         ,
        bdo_ready        => bdo_ready               ,
        key_update       => key_update              ,
        key_valid        => key_valid               ,
        key_ready        => key_ready               ,
        decrypt          => decrypt                 ,
        msg_auth_done    => msg_auth_done           ,        
        --! Control                                
        wr_nhkey         => wr_nhkey                ,
        is_decrypt       => is_decrypt              ,
        is_second        => is_second               ,
        start_hash       => start_hash              ,
        start_output     => start_output            ,
        output_done      => output_done             , 
        start_salsa      => start_salsa             ,
        salsa_init0      => salsa_init0             ,
        init_main_key    => init_main_key           ,
        init_ccha_key    => init_ccha_key           ,
        init_nhkey       => init_nhkey              ,
        init_poly_key    => init_poly_key           ,
        sel_bus_master   => sel_bus_master          ,
        sel_salsa_master => sel_salsa_master        ,
        sel_fifo         => sel_fifo                ,
        en_len           => en_len                  ,
        en_npub          => en_npub                 ,
        sel_tag          => sel_tag                 ,
        en_tag           => en_tag_sched            ,
        salsa_valid      => salsa_valid             ,
        salsa_ready      => sched_salsa_ready       ,
        hash_done        => hash_done       
    );
    
end structure;