-------------------------------------------------------------------------------
--! @file       HS1_Hash_Accum1.vhd
--! @brief      PSF Hash2
--!
--!             The first part of the accumulate unit.
--!             This unit determines the input and behave as the primary
--!             unit for all HS1_Hash module.
--!
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.HS1_SIV_pkg.all;

entity HS1_Hash_Accum1 is
    generic (
        G_MAX_LEN       : integer := 11
    );
    port    (
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! Input/Output
        bdi             : in  std_logic_vector(128              -1 downto 0);
        bdo             : out std_logic_vector(128              -1 downto 0);
        from_salsa      : in  std_logic_vector(128              -1 downto 0);
        data_o          : out std_logic_vector(64               -1 downto 0);
        --! Ext control
        salsa_valid     : in  std_logic;
        salsa_ready     : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_ready       : out std_logic;
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        fdi_valid       : in  std_logic;
        fdi_ready       : out std_logic;
        fdo_valid       : out std_logic;
        --! Control
        wr_key          : in  std_logic;
        start           : in  std_logic;
        start_output    : in  std_logic;
        output_done     : out std_logic;
        init_key        : in  std_logic;
        is_decrypt      : in  std_logic;
        is_second       : in  std_logic;
        len_a           : in  std_logic_vector(G_MAX_LEN        -1 downto 0);
        len_d           : in  std_logic_vector(G_MAX_LEN        -1 downto 0);
        en_tag          : out std_logic;
        oend            : out std_logic;
        ovalid          : out std_logic;
        oready          : in  std_logic
    );
end entity HS1_Hash_Accum1;

architecture structure of HS1_Hash_Accum1 is
    constant ZEROS      : std_logic_vector(128                  -1 downto 0)
                            := (others => '0');
    --! ======================================================================
    --! Datapath
    --! ======================================================================
    type t_nhkey is array (0 to 7) of std_logic_vector(127 downto 0);
    signal nhkey_ram    : t_nhkey;

    type t_data is array (0 to 3) of std_logic_vector(31 downto 0);
    signal input        : t_data;
    signal nhkey        : t_data;

    signal bdi_s        : std_logic_vector(128              -1 downto 0);
    signal bdo_s        : std_logic_vector(128              -1 downto 0);
    signal data         : std_logic_vector(128              -1 downto 0);
    signal len_data     : std_logic_vector(128              -1 downto 0);
    signal len_data_endian : std_logic_vector(128              -1 downto 0);
    signal from_keyram  : std_logic_vector(128              -1 downto 0);
    signal flag_vbits1  : std_logic_vector(128              -1 downto 0);
    signal flag_vbits2  : std_logic_vector(128              -1 downto 0);

    signal m1_a         : std_logic_vector(32               -1 downto 0);
    signal m1_b         : std_logic_vector(32               -1 downto 0);
    signal m1_c         : std_logic_vector(64               -1 downto 0);
    signal m2_a         : std_logic_vector(32               -1 downto 0);
    signal m2_b         : std_logic_vector(32               -1 downto 0);
    signal m2_c         : std_logic_vector(64               -1 downto 0);

    --! ======================================================================
    --! Controller
    --! ======================================================================
    signal len_ablk     : std_logic_vector(G_MAX_LEN-4      -1 downto 0);
    signal len_dreg     : std_logic_vector(G_MAX_LEN        -1 downto 0);
    signal st_vld       : std_logic_vector(4                -1 downto 0);
    signal st_end       : std_logic_vector(4                -1 downto 0);

    signal addr         : std_logic_vector(3                -1 downto 0);
    signal ctr          : std_logic_vector(2                -1 downto 0);

    --! Status
    signal flag_vbytes  : std_logic_vector(16               -1 downto 0);
    signal is_a_partial : std_logic;
    signal is_d_partial : std_logic;
    signal is_valid     : std_logic;
    signal is_end       : std_logic;
    signal is_ad        : std_logic;
    signal is_empty     : std_logic;

    --! Control
    signal set_ad       : std_logic;
    signal clr_ad       : std_logic;
    signal sel_data     : std_logic_vector(3                -1 downto 0);
    signal salsa_rdy    : std_logic;
    signal bdi_rdy      : std_logic;
    signal bdo_vld      : std_logic;
    signal en_len_a     : std_logic;
    signal en_len_d     : std_logic;
    signal en_mul       : std_logic;

    type t_state is (
        S_WAIT_START, S_INIT_KEY, S_STALL_READ, S_STALL,
        S_WAIT_AD, S_WAIT_DATA, S_LEN, S_WAIT_TAG,
        S_WAIT_SALSA_TAG, S_WAIT_LOAD,
        S_OUTPUT_ONLY_PREP, S_OUTPUT_ONLY);
    signal ps           : t_state;
    signal ns           : t_state;

    type t_vbytes_lookup is array (0 to 15) of
                          std_logic_vector(16               -1 downto 0);
    constant vbytes_lookup : t_vbytes_lookup :=
    (x"0000", x"8000", x"C000", x"E000", x"F000", x"F800", x"FC00", x"FE00",
     x"FF00", x"FF80", x"FFC0", x"FFE0", x"FFF0", x"FFF8", x"FFFC", x"FFFE");
begin
    len_data <= ZEROS(63 downto G_MAX_LEN) & len_a
               & ZEROS(63 downto G_MAX_LEN) & len_d;
    len_data_endian <= len_data(95 downto 64) & len_data(127 downto 96)
                       & len_data(31 downto 0) & len_data(63 downto 32);

    bdi_s <= switch_endian_byte(bdi, 32);
    gbdo:
    for i in 0 to 15 generate
        flag_vbits1(i*8+7 downto i*8) <= (others => flag_vbytes(i));
    end generate;
    flag_vbits2 <= switch_endian_byte(flag_vbits1, 32);
    bdo_s <= (from_salsa xor bdi_s) and flag_vbits2;
    bdo   <= switch_endian_byte(bdo_s, 32);

    with sel_data select
    data <= bdi_s           when "000",
            bdo_s           when "001",
            len_data_endian when "010",
            from_salsa      when "011",
            (others => '0') when others;

    from_keyram <= nhkey_ram(to_integer(unsigned(addr)));
    gdata:
    for i in 0 to 3 generate
        input(i) <= data(128-i*32-1 downto 128-(i+1)*32);
        nhkey(i) <= from_keyram(128-i*32-1 downto 128-(i+1)*32);
    end generate;

    m1_a <= std_logic_vector(unsigned(input(0)) + unsigned(nhkey(0)));
    m1_b <= std_logic_vector(unsigned(input(2)) + unsigned(nhkey(2)));
    m2_a <= std_logic_vector(unsigned(input(1)) + unsigned(nhkey(1)));
    m2_b <= std_logic_vector(unsigned(input(3)) + unsigned(nhkey(3)));

    mult1: entity work.Mult32_PPL
    port map (
        clk => clk,
        aa  => m1_a,
        bb  => m1_b,
        cc  => m1_c,
        en  => en_mul
    );
    mult2: entity work.Mult32_PPL
    port map (
        clk => clk,
        aa  => m2_a,
        bb  => m2_b,
        cc  => m2_c,
        en  => en_mul
    );

    data_o <= std_logic_vector(unsigned(m1_c) + unsigned(m2_c));

    is_a_partial <= len_a(3) or len_a(2) or len_a(1) or len_a(0);
    is_d_partial <= len_d(3) or len_d(2) or len_d(1) or len_d(0);

    ovalid <= st_vld(0);
    oend   <= st_end(0);
    process(clk)
    begin
        if rising_edge(clk) then
            --! =================
            --! Datapath
            --! =================
            if (wr_key = '1') then
                nhkey_ram(to_integer(unsigned(addr))) <= from_salsa;
            end if;

            --! =================
            --! Controller
            --! =================
            if (rst = '1') then
                st_vld <= (others => '0');
            elsif (en_mul = '1') then
                st_vld(3) <= is_valid;
                st_vld(2) <= st_vld(3);
                st_vld(1) <= st_vld(2);
                st_vld(0) <= st_vld(1);

                st_end(3) <= is_end;
                st_end(2) <= st_end(3);
                st_end(1) <= st_end(2);
                st_end(0) <= st_end(1);
            end if;

            if (set_ad = '1') then
                is_ad <= '1';
            elsif (clr_ad = '1') then
                is_ad <= '0';
            end if;
            --! Keep track of start for 4 cycles
            if (start = '1') then
                len_ablk <= std_logic_vector(unsigned(
                            len_a(G_MAX_LEN-1 downto 4))
                            + (is_a_partial & ""));
            elsif (en_len_a = '1') then
                len_ablk <= std_logic_vector(unsigned(len_ablk)-1);
            end if;

            if ((ps = S_STALL and unsigned(ctr) = 3) or ps = S_STALL_READ) then
                if (set_ad = '1' or is_ad = '1') then
                    if (unsigned(len_ablk) = 0) then
                        if (unsigned(addr) = 6) then
                            if (unsigned(len_dreg) = 0) then
                                is_empty <= '1';
                            else
                                is_empty <= '0';
                            end if;
                        else
                            is_empty <= '1';
                        end if;
                    else
                        is_empty <= '0';
                    end if;
                elsif (is_ad = '0') then
                    if (unsigned(len_dreg) = 0) then
                        is_empty <= '1';
                    else
                        is_empty <= '0';
                    end if;
                end if;
            end if;

            if (start = '1') then
                len_dreg <= len_d;
            elsif (en_len_d = '1') then
                if (unsigned(len_dreg(G_MAX_LEN-1 downto 4)) > 0) then
                    --! Full block
                    flag_vbytes <= (others => '1');
                else
                    flag_vbytes <= vbytes_lookup(
                                       to_integer(
                                       unsigned(len_dreg(3 downto 0))));
                end if;
                if (unsigned(len_dreg(G_MAX_LEN-1 downto 4)) > 0) then
                    len_dreg(G_MAX_LEN-1 downto 4) <= std_logic_vector(unsigned(len_dreg(G_MAX_LEN-1 downto 4))-1);
                else
                    len_dreg(3 downto 0) <= (others => '0');
                end if;
            end if;

            if (start = '1' or init_key = '1') then
                ctr <= (others => '0');
            elsif (en_mul = '1' and is_valid = '1') then
                ctr <= std_logic_vector(unsigned(ctr) + 1);
            end if;

            if (start = '1' or init_key = '1') then
                addr <= (others => '0');
            elsif (en_mul = '1' and (is_valid = '1' or wr_key = '1')) then
                if (unsigned(ctr) = 3) then
                    if (unsigned(addr) = 6) then
                        addr <= (others => '0');
                    else
                        addr <= std_logic_vector(unsigned(addr) - 2);
                    end if;
                else
                    addr <= std_logic_vector(unsigned(addr) + 1);
                end if;
            end if;

            if (rst = '1') then
                ps <= S_WAIT_START;
            else
                ps    <= ns;
            end if;
        end if;
    end process;

    pcomb:
    process(ps, len_ablk, len_dreg, start, start_output, ctr, wr_key,
        is_empty, is_ad, fdi_valid, init_key,
        salsa_valid, bdi_valid, bdo_ready, oready, st_vld)
    begin
        ns              <= ps;
        bdi_rdy         <= '0';
        bdo_vld         <= '0';
        salsa_rdy       <= '0';
        sel_data        <= "000";
        en_len_a        <= '0';
        en_len_d        <= '0';
        en_mul          <= '0';
        en_tag          <= '0';
        set_ad          <= '0';
        clr_ad          <= '0';
        fdi_ready       <= '0';
        fdo_valid       <= '0';
        is_valid        <= '0';
        is_end          <= '0';
        output_done     <= '0';


        if (ps = S_WAIT_START) then
            if (init_key = '1') then
                ns <= S_INIT_KEY;
            elsif (start = '1') then
                ns <= S_STALL_READ;
            elsif (start_output = '1') then
                ns <= S_OUTPUT_ONLY_PREP;
            end if;
        elsif (ps = S_INIT_KEY) then
            if (wr_key = '1') then
                en_mul <= '1';
                if (unsigned(addr) = 6) then
                    ns <= S_WAIT_START;
                end if;
            end if;
        elsif (ps = S_STALL_READ) then
            if (is_second = '0' and is_decrypt = '0')
                or (is_second = '1' and is_decrypt = '1')
            then
                if (unsigned(len_ablk) /= 0) then
                    set_ad <= '1';
                    en_len_a <= '1';
                    ns <= S_WAIT_AD;
                elsif (unsigned(len_dreg) /= 0) then
                    clr_ad   <= '1';
                    en_len_d <= '1';
                    ns <= S_WAIT_DATA;
                else
                    clr_ad <= '1';
                    ns <= S_LEN;
                end if;
            elsif (is_second = '1' and is_decrypt = '0') then
                clr_ad <= '1';
                ns     <= S_WAIT_SALSA_TAG;
            else
                ns     <= S_WAIT_TAG;
            end if;
        elsif (ps = S_STALL) then
            if (oready = '1') then
                en_mul <= '1';
                if (unsigned(ctr) = 3) then
                    if (is_ad = '1') then
                        if (is_empty = '0') then
                            bdi_rdy  <= '1';
                        end if;
                    else
                        bdi_rdy      <= '1';
                        if (is_decrypt = '1') then
                            bdo_vld   <= '1';
                            salsa_rdy <= '1';
                        else
                            fdo_valid <= '1';
                        end if;
                    end if;
                    if (unsigned(len_ablk) /= 0
                        or (is_ad = '1' and unsigned(addr) /= 6))
                    then
                        if (unsigned(len_ablk) /= 0) then
                            en_len_a <= '1';
                        end if;
                        ns <= S_WAIT_AD;
                    else
                        clr_ad <= '1';
                        if (unsigned(len_dreg) /= 0) then
                            en_len_d <= '1';
                            ns <= S_WAIT_DATA;
                        else
                            ns <= S_LEN;
                        end if;
                    end if;
                end if;
            end if;
            is_valid <= '1';
            if (is_ad = '1') then
                if (is_empty = '1') then
                    sel_data <= "111";
                end if;
            else
                if (is_decrypt = '1') then
                    sel_data <= "001";
                end if;
            end if;
        elsif (ps = S_WAIT_AD) then
            if (is_empty = '1') then
                sel_data <= "111";
            end if;
            if (oready = '1') then
                en_mul   <= '1';
                if ((is_empty = '0' and bdi_valid = '1')
                        or (is_empty = '1'))
                then
                    is_valid <= '1';
                    ns <= S_STALL;
                end if;
            end if;
        elsif (ps = S_WAIT_DATA) then
            if (is_decrypt = '1') then
                sel_data <= "001";
            end if;
            if (oready = '1') then
                if ((is_decrypt = '1' and salsa_valid = '1' and bdo_ready = '1')
                        or (is_decrypt = '0' and bdi_valid = '1'))
                then
                    is_valid <= '1';
                    ns       <= S_STALL;
                end if;
                en_mul   <= '1';
            end if;
        elsif (ps = S_LEN) then
            sel_data <= "010";
            is_valid <= '1';
            is_end   <= '1';
            if (oready = '1') then
                en_mul   <= '1';
                if (unsigned(ctr) = 3) then
                    ns <= S_WAIT_LOAD;
                end if;
            end if;
        elsif (ps = S_WAIT_TAG) then
            is_end   <= '1';
            if (oready = '1') then
                en_mul   <= '1';
                if (bdi_valid = '1') then
                    is_valid <= '1';
                    if (unsigned(ctr) = 3) then
                        en_tag   <= '1'; --! Write tag to register
                        bdi_rdy  <= '1';
                        ns       <= S_WAIT_LOAD;
                    end if;
                end if;
            end if;
        elsif (ps = S_WAIT_SALSA_TAG) then
            sel_data <= "011";
            is_valid <= '1';
            is_end   <= '1';
            if (oready = '1') then
                en_mul   <= '1';
                if (unsigned(ctr) = 3) then
                    salsa_rdy <= '1';
                    ns <= S_WAIT_LOAD;
                end if;
            end if;
        elsif (ps = S_WAIT_LOAD) then
            if (oready = '1') then
                en_mul <= '1';
                if (st_vld(1) = '0' and st_vld(0) = '1') then
                    ns <= S_WAIT_START;
                end if;
            end if;
        elsif (ps = S_OUTPUT_ONLY_PREP) then
            en_len_d <= '1';
            ns <= S_OUTPUT_ONLY;
        elsif (ps = S_OUTPUT_ONLY) then
            if (fdi_valid = '1') then
                if (salsa_valid = '1') then
                    salsa_rdy    <= '1';
                    fdi_ready <= '1';
                    bdo_vld      <= '1';
                    en_len_d     <= '1';
                end if;
            else
                output_done  <= '1';
                ns <= S_WAIT_START;
            end if;
        end if;

    end process;
    bdi_ready <= bdi_rdy;
    bdo_valid <= bdo_vld;
    salsa_ready <= salsa_rdy;

end architecture structure;