-------------------------------------------------------------------------------
--! @file       HS1_Hash_Accum2.vhd
--! @brief      PSF Hash2 Accumulate unit for 'a0' and 'a1' intermediate values
--!
--!             Data arrive in 64-bit word for a total of 256-bit
--!             (2 PSF_Hash calls)
--!
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity HS1_Hash_Accum2 is
    port    (
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! Input/Output
        data_i          : in  std_logic_vector(60            -1 downto 0);
        data_o          : out std_logic_vector(60            -1 downto 0);
        --! Control
        start           : in  std_logic;
        iend            : in  std_logic;
        ivalid          : in  std_logic;
        iready          : out std_logic;
        oend            : out std_logic;
        ovalid          : out std_logic;
        oready          : in  std_logic
    );
end entity HS1_Hash_Accum2;

architecture structure of HS1_Hash_Accum2 is
    --! ======================================================================
    --! Datapath
    --! ======================================================================
    type t_accum is array (0 to 3) of std_logic_vector(59 downto 0);
    signal accum : t_accum;

    signal to_accum_add : std_logic_vector(60               -1 downto 0);
    signal to_accum     : std_logic_vector(60               -1 downto 0);
    
    --! ======================================================================
    --! Controller
    --! ======================================================================
    --! 4 bit - 2-bit for counting upto 512-bit data processing
    --!         Note: Each 128-bit takes 4 cycles
    --!               (process 256-bit accum, lower bits)
    --!       - 2-bit for counting 4 64-bit accum blocks
    signal ctr          : std_logic_vector( 4               -1 downto 0);
    
    --! Status
    signal is_process   : std_logic;
    signal is_init      : std_logic;
    
    --! Control
    signal irdy         : std_logic;
    signal ovld         : std_logic;
    signal sel_init     : std_logic;
    signal en_accum     : std_logic;
begin

    to_accum_add <= (others => '0') when sel_init = '1' else accum(0);
    to_accum     <= std_logic_vector(unsigned(to_accum_add) + unsigned(data_i));
    data_o       <= to_accum;

    process(clk)
    begin
        if rising_edge(clk) then
            --! =================
            --! Datapath
            --! =================
            if (en_accum = '1') then
                accum(3) <= to_accum;
                accum(2) <= accum(3);
                accum(1) <= accum(2);
                accum(0) <= accum(1);
            end if;
            
            --! =================
            --! Controller
            --! =================
                --! Keep track of start for 4 cycles
            if (rst = '1' or start = '1') then
                is_init <= '1';
            elsif (ctr(1 downto 0) = "11") then
                is_init <= '0';
            end if;

            --! Default values
            if (rst = '1') 
                or (en_accum = '1' and iend = '1' 
                    and ctr(1 downto 0) = "11") 
            then
                is_process <= '0';
            elsif (start = '1') then
                is_process <= '1';
            end if;
            
            if (start = '1') then
                ctr <= (others => '0');
            elsif (en_accum = '1') then
                ctr <= std_logic_vector(unsigned(ctr) + 1);
            end if;
        end if;
    end process;
    sel_init <= '1' when ctr(3 downto 2) = "00" else '0';
    en_accum <= irdy and ivalid;
    irdy     <= is_process and (not ovld or (ovld and oready));
    oend     <= iend;
    ovld     <= '1' when (ctr(3 downto 2) = "11")
                         or (iend = '1' and ivalid = '1')
                    else '0';
    iready   <= irdy;
    ovalid   <= ovld;
end architecture structure;