-------------------------------------------------------------------------------
--! @file       HS1_Hash_Sum.vhd
--! @brief      HS1 summation unit.
--!
--!             This module performs poly_step and poly_final.
--!
--!             Key Initialization:
--!             Load PolyKey in 4 consecutive clock cycles.
--!             Assumes key is always available once started.
--!
--!             Operation:
--!             Poly step is performed as soon as data is available.
--!             Poly final is performed at the end. Once.
--!
--!
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity HS1_Hash_Sum is
    port    (
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! Input/Output
        key             : in  std_logic_vector(128           -1 downto 0);
        data_i          : in  std_logic_vector(60            -1 downto 0);
        hash            : out std_logic_vector(256           -1 downto 0);
        --! Control
        start           : in  std_logic;
        init_pkey       : in  std_logic;
        init_ckey       : in  std_logic;
        iend            : in  std_logic;
        ivalid          : in  std_logic;
        iready          : out std_logic;
        done            : out std_logic
    );
end entity HS1_Hash_Sum;

architecture structure of HS1_Hash_Sum is
    constant ZEROS     : std_logic_vector(128           -1 downto 0)
                            := (others => '0');
    constant ONES      : std_logic_vector(128           -1 downto 0)
                            := (others => '1');
    --! ======================================================================
    --! Datapath
    --! ======================================================================
    type t_sr      is array (0 to 3) of std_logic_vector(63 downto 0);
    type t_sr_poly is array (0 to 3) of std_logic_vector(59 downto 0);
    signal sum         : t_sr;
    signal poly_key    : t_sr_poly;
    signal ccha_key    : t_sr;

    signal to_key_mux  : std_logic_vector(64               -1 downto 0);
    signal to_poly_key : std_logic_vector(60               -1 downto 0);
    signal to_ccha_key : std_logic_vector(64               -1 downto 0);
    signal to_sum      : std_logic_vector(64               -1 downto 0);
    signal key_multh   : std_logic_vector(32               -1 downto 0);
    signal key_multl   : std_logic_vector(32               -1 downto 0);
    signal pmul_h      : std_logic_vector(64               -1 downto 0);
    signal pmul_l      : std_logic_vector(64               -1 downto 0);
    signal pmul_m      : std_logic_vector(64               -1 downto 0);
    signal to_hash     : std_logic_vector(64               -1 downto 0);
    signal from_pmul   : std_logic_vector(64               -1 downto 0);
    signal from_pstep  : std_logic_vector(64               -1 downto 0);
    signal from_pfinal : std_logic_vector(64               -1 downto 0);

    --! ======================================================================
    --! Controller
    --! ======================================================================
    --! 4 bit - 2-bit for counting upto 512-bit data processing
    --!         Note   : Each 128-bit takes 4 cycles
    --!               (process 256-bit accum, lower bits)
    --!       - 2-bit for counting 4 64-bit accum blocks
    signal ctr         : std_logic_vector( 4                -1 downto 0);

    --! Control
    signal sel_sum     : std_logic_vector( 2                -1 downto 0);
    signal sel_key     : std_logic;
    signal ld_ctr      : std_logic;
    signal en_mul      : std_logic;
    signal en_ctr      : std_logic;
    signal en_key_poly : std_logic;
    signal en_key_ccha : std_logic;
    signal en_sum      : std_logic;

    type t_state is (S_WAIT_START, S_INIT_SUM, S_INIT_CCHA_KEY,
        S_INIT_POLY_KEY, S_PROCESS, S_LOAD_ACCUM, S_POLY_FINAL);
    signal ps           : t_state;
    signal ns           : t_state;
begin

    with sel_sum select
    to_sum <= from_pmul                                       when "00",
              from_pstep                                      when "01",
              std_logic_vector(to_unsigned(1, to_sum'length)) when others;

    --! =======================================================================
    --! Key logic
    --! =======================================================================
    to_key_mux <= key(31 downto 0) & key(63 downto 32)
                  when ctr(0) = '1'
                  else key(95 downto 64) & key(127 downto 96);
    to_poly_key <= to_key_mux(59 downto 0)
                   when sel_key = '1' else poly_key(0);
    to_ccha_key <= to_key_mux when sel_key = '1' else ccha_key(0);

    --! =======================================================================
    --! Multiplication logic
    --! =======================================================================
    key_multh <= x"0" & poly_key(0)(59 downto 32)
                 when ctr(0) = '1'
                 else poly_key(0)(31 downto 0);
    key_multl <= x"0" & poly_key(0)(59 downto 32)
                 when ctr(0) = '0'
                 else poly_key(0)(31 downto 0);

    multH: entity work.Mult32_PPL
    port map (
        clk => clk,
        aa  => sum(0)(63 downto 32),
        bb  => key_multh,
        cc  => pmul_h,
        en  => en_mul
    );
    multL: entity work.Mult32_PPL
    port map (
        clk => clk,
        aa  => sum(0)(31 downto  0),
        bb  => key_multl,
        cc  => pmul_l,
        en  => en_mul
    );

    --! =======================================================================
    --! Poly Step logic
    --! =======================================================================
    pmulBlock: block
        --! Last part of pmul and pre-calculate last part of poly-step
        --! ====
        --! pmul64 {
        --!     ...
        --!     h += (m >> 32); l += (m << 32);
        --!     if (l < (m << 32)) h += 1;
        --!     *dl=l; *dh=h;
        --! }
        --!
        --! Result = (l & m61) + ((h << 3) | (l >> 61))

        signal pmul_m_hi : unsigned(63 downto 0);
        signal tmp_l     : unsigned(63 downto 0);
        signal comp      : unsigned( 0 downto 0);
        signal tmp_h     : unsigned(63 downto 0);
        signal tmp1      : unsigned(63 downto 0);
        signal tmp2      : std_logic_vector(63 downto 0);
        signal tmp3      : std_logic_vector(63 downto 0);
        signal tmp4      : unsigned(63 downto 0);
        signal result    : unsigned(63 downto 0);
    begin
        pmul_m_hi <= unsigned(pmul_m(31 downto 0)) & x"00000000";
        tmp_l     <= unsigned(pmul_l) + pmul_m_hi;
        comp      <= "1" when tmp_l < pmul_m_hi else "0";

        tmp_h     <= unsigned(pmul_h) + unsigned(pmul_m(63 downto 32)) + comp;
        --! (l & m61)
        tmp1  <= "000" & tmp_l(60 downto 0);
        --!     (h << 3)
        tmp2  <= std_logic_vector(tmp_h(60 downto 0) & "000");
        --!     (l >> 61)
        tmp3  <= ZEROS(60 downto 0) & std_logic_vector(tmp_l(63 downto 61));
        --! ((h << 3) | (l >> 61))
        tmp4  <= unsigned(tmp2 or tmp3);
        result    <= tmp1 + tmp4;

        from_pmul <= std_logic_vector(result);
    end block;


    --! Last part of poly step
    --! poly_step {
    --!     ...
    --!     return (l & m61) + ((h << 3) | (l >> 61)) + m;
    --!     # = return from_pmul + m;
    --! }
    from_pstep <= std_logic_vector(unsigned(sum(0)) + unsigned(data_i));

    --! =======================================================================
    --! Polyfinal + Hash logic
    --! =======================================================================
    polyFinalBlock: block
        --! poly_finalize {
        --!     a = (a & m61) + (a >> 61);
        --!     if (a == m61) a = 0;
        --!     return a;
        --! }
        signal from_sum : std_logic_vector(63 downto 0);
        signal to_tmp : std_logic_vector(63 downto 0);
        signal tmp_a  : unsigned(61 downto 0);
    begin
        from_sum <= sum(0);
        tmp_a    <= ('0' & unsigned(from_sum(60 downto 0)))
                    + unsigned(from_sum(63 downto 61));
        from_pfinal <= (others => '0')
                        when tmp_a(60 downto 0) = unsigned(ONES(60 downto 0))
                        else "00" & std_logic_vector(tmp_a);
    end block;
    to_hash <= from_pfinal xor ccha_key(0);

    preg:
    process(clk)
    begin
        if rising_edge(clk) then
            --! =================
            --! Datapath
            --! =================
            if (en_sum = '1') then
                sum(3) <= to_sum;
                sum(2) <= sum(3);
                sum(1) <= sum(2);
                sum(0) <= sum(1);
            end if;

            if (en_key_poly = '1') then
                poly_key(3) <= to_poly_key;
                poly_key(2) <= poly_key(3);
                poly_key(1) <= poly_key(2);
                poly_key(0) <= poly_key(1);
            end if;

            if (en_key_ccha = '1') then
                ccha_key(3) <= to_ccha_key;
                ccha_key(2) <= ccha_key(3);
                ccha_key(1) <= ccha_key(2);
                ccha_key(0) <= ccha_key(1);
            end if;

            if (ps = S_POLY_FINAL) then
                if (ctr(1 downto 0) = "00") then
                    hash(255 downto 192) <= to_hash;
                end if;
                if (ctr(1 downto 0) = "01") then
                    hash(191 downto 128) <= to_hash;
                end if;
                if (ctr(1 downto 0) = "10") then
                    hash(127 downto  64) <= to_hash;
                end if;
                if (ctr(1 downto 0) = "11") then
                    hash( 63 downto   0) <= to_hash;
                end if;
            end if;

            if (en_mul = '1') then
                pmul_m <= std_logic_vector(unsigned(pmul_h) + unsigned(pmul_l));
            end if;
            --! =================
            --! Controller
            --! =================
            if (rst = '1') then
                ps <= S_WAIT_START;
            else
                ps <= ns;
            end if;

            if (ld_ctr = '1') then
                ctr <= (others => '0');
            elsif (en_ctr = '1') then
                ctr <= std_logic_vector(unsigned(ctr) + 1);
            end if;

            if (rst = '1' or (ld_ctr = '1' and ps = S_LOAD_ACCUM)) then
                iready <= '0';
            elsif (ld_ctr = '1' and ps = S_PROCESS) then
                iready <= '1';
            end if;
        end if;
    end process;

    pcomb:
    process(ps, ivalid, ctr, iend, start, init_pkey, init_ckey)
    begin
        ns <= ps;
        sel_sum <= "00";
        sel_key <= '0';
        ld_ctr <= '0';
        en_ctr <= '0';
        en_key_poly <= '0';
        en_key_ccha <= '0';
        en_sum <= '0';
        en_mul <= '0';
        done   <= '0';

        if (ps = S_WAIT_START) then
            ld_ctr <= '1';
            if (start = '1') then
                ns <= S_INIT_SUM;
            elsif (init_pkey = '1') then
                ns <= S_INIT_POLY_KEY;
            elsif (init_ckey = '1') then
                ns <= S_INIT_CCHA_KEY;
            end if;
        elsif (ps = S_INIT_SUM) then
            sel_sum <= "10";
            en_sum <= '1';
            en_ctr <= '1';
            if (unsigned(ctr) = 3) then
                ld_ctr <= '1';
                ns     <= S_PROCESS;
            end if;
        elsif (ps = S_INIT_POLY_KEY) then
            en_ctr      <= '1';
            sel_key     <= '1';
            en_key_poly <= '1';
            if (unsigned(ctr) = 3) then
                ld_ctr <= '1';
                ns     <= S_WAIT_START;
            end if;
        elsif (ps = S_INIT_CCHA_KEY) then
            en_ctr      <= '1';
            sel_key     <= '1';
            en_key_ccha <= '1';
            if (unsigned(ctr) = 3) then
                ld_ctr <= '1';
                ns     <= S_WAIT_START;
            end if;
        elsif (ps = S_PROCESS) then
            --! Perform a pre-calculation of poly_step up until before the end
            --! of poly_step. This is everything before the addition with
            --! data from accum (m input in C code).
            en_ctr <= '1';
            
            en_mul <= '1';
            if (unsigned(ctr) < 8 and ctr(0) = '1') then
                en_key_poly <= '1';
            end if;
            if (ctr(0) = '1') then
                en_sum <= '1';
            end if;
            if (unsigned(ctr) = 11) then
                ld_ctr <= '1';
                ns     <= S_LOAD_ACCUM;
            end if;
        elsif (ps = S_LOAD_ACCUM) then
            sel_sum <= "01";
            if (ivalid = '1') then
                en_sum <= '1';
                en_ctr <= '1';
                if (unsigned(ctr) = 3) then
                    ld_ctr <= '1';
                    if (iend = '1') then
                        ns <= S_POLY_FINAL;
                    else
                        ns <= S_PROCESS;
                    end if;
                end if;
            end if;
        elsif (ps = S_POLY_FINAL) then
            en_ctr      <= '1';
            en_key_ccha <= '1';
            en_sum      <= '1';
            sel_sum     <= "10";
            if (unsigned(ctr) = 3) then
                ld_ctr <= '1';
                done   <= '1';
                ns     <= S_WAIT_START;
            end if;
        end if;
    end process;
end architecture structure;