-------------------------------------------------------------------------------
--! @file       HS1_SIV_Salsa20.vhd
--! @brief      A wrapper for Salsa20 used by HS1_SIV
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.HS1_SIV_pkg.all;

entity HS1_SIV_Salsa20 is
    generic (
        G_MAX_LEN    : integer := 11
    );
    port    (
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! Input/out_r
        key             : in  std_logic_vector(128           -1 downto 0);
        npub            : in  std_logic_vector( 96           -1 downto 0);
        hash            : in  std_logic_vector(256           -1 downto 0);
        yy              : out std_logic_vector(128           -1 downto 0);
        --! Control
        start           : in  std_logic;
        init0           : in  std_logic;
        init_key        : in  std_logic;
        busy            : out std_logic;
        ovalid          : out std_logic;
        oready          : in  std_logic
    );
end entity HS1_SIV_Salsa20;

architecture structure of HS1_SIV_Salsa20 is    
    --! ======================================================================
    --! Datapath
    --! ======================================================================
    signal to_salsa     : std_logic_vector(512              -1 downto 0);
    signal from_salsa   : std_logic_vector(512              -1 downto 0);
    signal from_iv_sel  : std_logic_vector( 96              -1 downto 0);
    signal key_ext      : std_logic_vector(256              -1 downto 0);
    signal key_or_hash  : std_logic_vector(256              -1 downto 0);
    signal to_out_r     : std_logic_vector(512              -1 downto 0);
    signal out_r        : std_logic_vector(512              -1 downto 0);
    signal new_ctr      : std_logic_vector(G_MAX_LEN-4      -1 downto 0);
    signal ctr          : std_logic_vector(G_MAX_LEN-4      -1 downto 0);
    signal ctr_r        : std_logic_vector(G_MAX_LEN-4      -1 downto 0);

    --! ======================================================================
    --! Controller
    --! ======================================================================
    signal remblk       : std_logic_vector(3                -1 downto 0);
    signal init_key_r   : std_logic;
    signal sstart       : std_logic;
    signal sdone        : std_logic;
    signal score_rdy    : std_logic;

    signal is_empty     : std_logic;
    signal is_init      : std_logic;

    type t_state is (S_WAIT_READ, S_WAIT_DONE);
    signal ps           : t_state;
    signal ns           : t_state;
begin
    new_ctr      <= std_logic_vector(unsigned(ctr) + natural(1));
    key_ext      <= key & key;
    key_or_hash  <= switch_endian_byte(key_ext, 32)
                    when init_key_r = '1' 
                    else hash(223 downto 192) & hash(255 downto 224) 
                         & hash(159 downto 128) & hash(191 downto 160)
                         & hash( 95 downto  64) & hash(127 downto  96)
                         & hash( 31 downto   0) & hash( 63 downto  32);
    from_iv_sel  <= CONS_STR    
                    when init_key_r = '1' 
                    else switch_endian_byte(npub, 32);
    to_salsa     <= SIGMA & key_or_hash & ZEROS(31 downto G_MAX_LEN-4)
                    & ctr   & from_iv_sel;
    to_out_r     <= SIGMA & key_or_hash & ZEROS(31 downto G_MAX_LEN-4)
                    & ctr_r & from_iv_sel;

    u_salsa: entity work.Salsa20
    port map (
        clk     => clk,         rst     => rst,
        start   => sstart,      ready   => score_rdy,  done => sdone,
        xx      => to_salsa,    yy      => from_salsa
    );

    is_empty <= '1' when unsigned(remblk) = 0
                         or (unsigned(remblk) = 1 and oready = '1')
                    else '0';

    with remblk(1 downto 0) select
    yy <=   out_r(512-128*0-1 downto 512-128*1) when "00",
            out_r(512-128*1-1 downto 512-128*2) when "11",
            out_r(512-128*2-1 downto 512-128*3) when "10",
            out_r(512-128*3-1 downto 512-128*4) when others;
    ovalid <= '1' when unsigned(remblk) > 0 else '0';

    preg:
    process(clk)
    begin
        if rising_edge(clk) then
            --! =================
            --! Datapath
            --! =================
            if (sstart = '1') then
                ctr_r  <= ctr;
                for i in 0 to 15 loop
                    out_r(i*32+31 downto i*32) <= std_logic_vector(
                        unsigned(from_salsa(i*32+31 downto i*32)) + 
                        unsigned(to_out_r(i*32+31 downto i*32)));
                end loop;
            end if;

            if (start = '1') then
                if (init0 = '1') then
                    ctr <= (others => '0');
                else
                    ctr <= std_logic_vector(to_unsigned(1, ctr'length));
                end if;
            elsif (sstart = '1') then
                ctr <= new_ctr;
            end if;


            --! =================
            --! Controller
            --! =================
            if (start = '1') then
                init_key_r <= init_key;
            end if;

            if (start = '1') then
                is_init <= '1';
            elsif (sdone = '1') then
                is_init <= '0';
            end if;

            if (rst = '1' or start = '1') then
                remblk <= "000";
            elsif (is_empty = '1'
                and ((ps = S_WAIT_DONE and sdone = '1')
                     or (ps = S_WAIT_READ and is_init = '0')))
            then
                remblk <= "100";
            elsif (unsigned(remblk) /= 0 and oready = '1') then
                remblk <= std_logic_vector(unsigned(remblk) - 1);
            end if;

            if (rst = '1' or start = '1') then
                ps <= S_WAIT_READ;
            else
                ps <= ns;
            end if;
        end if;
    end process;

    pcomb:
    process(ps, is_empty, sdone, score_rdy)
    begin
        ns     <= ps;
        sstart <= '0';
        busy   <= '0';
        if (ps = S_WAIT_READ) then
            if (is_empty = '1' and score_rdy = '1') then
                sstart <= '1';
                ns <= S_WAIT_DONE;
            end if;
        elsif (ps = S_WAIT_DONE) then
            busy <= '1';
            if (sdone = '1') then
                if (is_empty = '1' and score_rdy = '1') then
                    sstart <= '1';
                else
                    ns     <= S_WAIT_READ;
                end if;
            end if;
        end if;
    end process;
end architecture structure;