-------------------------------------------------------------------------------
--! @file       HS1_SIV_Sched.vhd
--! @brief      A scheduler for HS1-SIV.
--!
--!             This module controls the overall operation of
--!             HS1-SIV submodules, and provide sub-modules with bus-mastering
--!             capability.
--!
--!             Enc:
--!                 Hash -> Salsa(1) -> Hash(1) -> Salsa
--!             Dec:
--!                 Hash(1) -> Hash/Salsa
--!
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity HS1_SIV_Sched is
    port    (
        clk                 : in  std_logic;
        rst                 : in  std_logic;
        --! Ext. Control
        bdi_valid           : in  std_logic;
        bdi_ready           : out std_logic;
        bdo_valid           : out std_logic;
        bdo_ready           : in  std_logic;
        key_update          : in  std_logic;
        key_valid           : in  std_logic;
        key_ready           : out std_logic;
        decrypt             : in  std_logic;
        msg_auth_done       : out std_logic;
        --! Control
        wr_nhkey            : out std_logic;
        is_decrypt          : out std_logic;
        is_second           : out std_logic;
        start_hash          : out std_logic;
        start_output        : out std_logic;
        start_salsa         : out std_logic;
        salsa_init0         : out std_logic;
        init_main_key       : out std_logic;
        init_ccha_key       : out std_logic;
        init_nhkey          : out std_logic;
        init_poly_key       : out std_logic;

        sel_bus_master      : out std_logic;
        sel_salsa_master    : out std_logic;
        sel_fifo            : out std_logic;
        en_len              : out std_logic;
        en_npub             : out std_logic;
        sel_tag             : out std_logic;
        en_tag              : out std_logic;

        salsa_valid         : in  std_logic;
        salsa_ready         : out std_logic;
        output_done         : in  std_logic;
        hash_done           : in  std_logic
    );
end entity HS1_SIV_Sched;

architecture structure of HS1_SIV_Sched is
    signal ctr          : std_logic_vector(2 downto 0);
    signal set_second   : std_logic;
    signal is_decrypt_r : std_logic;
    signal is_second_r  : std_logic;
    signal ld_ctr       : std_logic;
    signal en_ctr       : std_logic;
    signal st_hash      : std_logic;
    signal st_salsa     : std_logic;
    signal st_output    : std_logic;

    type t_state is (S_WAIT, S_INIT_KEY,
        S_WAIT_INIT_CKEY, S_WAIT_INIT_NHKEY, S_WAIT_INIT_PKEY,
        S_LD_LEN, S_LD_TAG, S_LD_NPUB,
        S_WAIT_HASH, S_WAIT_SALSA_OUT, S_WAIT_SALSA_TAG,
        S_WR_TAG
        );
    signal ps           : t_state;
    signal ns           : t_state;
begin


    preg:
    process(clk)
    begin
        if rising_edge(clk) then
            start_hash    <= st_hash;
            start_output  <= st_output;

            if (rst = '1') then
                ps <= S_WAIT;
            else
                ps <= ns;
            end if;
            if (ps = S_WAIT) then
                is_decrypt_r <= decrypt;
            end if;

            if (ps = S_WAIT) then
                is_second_r <= '0';
            elsif (set_second = '1') then
                is_second_r <= '1';
            end if;

            if (ld_ctr = '1') then
                ctr <= (others => '0');
            elsif (en_ctr = '1') then
                ctr <= std_logic_vector(unsigned(ctr) + 1);
            end if;
        end if;
    end process;
    is_decrypt  <= is_decrypt_r;
    is_second   <= is_second_r;
    start_salsa <= st_salsa;
    salsa_init0 <= '1' when (is_second_r = '0' and is_decrypt_r = '0')
                    or (is_second_r = '1' and is_decrypt_r = '1')
                    or (ps = S_WAIT) else '0';


    pcomb:
    process(ps, bdi_valid, bdo_ready,
        hash_done, salsa_valid, output_done,
        is_decrypt_r, is_second_r,
        key_update, key_valid, ctr)
    begin
        ns              <= ps;
        ld_ctr          <= '0';
        en_ctr          <= '0';
        wr_nhkey        <= '0';
        sel_salsa_master  <= '0';
        sel_bus_master  <= '0';

        en_len          <= '0';
        en_npub         <= '0';
        en_tag          <= '0';
        sel_tag         <= '0';
        set_second      <= '0';

        salsa_ready     <= '0';
        key_ready       <= '0';

        msg_auth_done   <= '0';
        bdi_ready       <= '0';
        bdo_valid       <= '0';
        st_hash         <= '0';
        st_output    <= '0';
        st_salsa        <= '0';
        sel_fifo        <= '0';
        init_main_key   <= '0';
        init_ccha_key   <= '0';
        init_nhkey      <= '0';
        init_poly_key   <= '0';

        if (ps = S_WAIT) then
            ld_ctr  <= '1';
            if (key_valid = '1' and key_update = '1') then
                ns            <= S_INIT_KEY;
                st_salsa      <= '1';
                init_main_key <= '1';
            elsif (bdi_valid = '1' and key_update = '0') then
                ns            <= S_LD_LEN;
            end if;
        elsif (ps = S_INIT_KEY) then
            init_nhkey <= '1';
            if (salsa_valid = '1') then
                init_ccha_key <= '1';
                ns <= S_WAIT_INIT_CKEY;
            end if;
        elsif (ps = S_WAIT_INIT_CKEY) then
            sel_salsa_master <= '1';
            en_ctr <= '1';
            if (ctr(0) = '1') then
                salsa_ready <= '1';
            end if;
            if (ctr(1 downto 0) = "11") then
                ld_ctr <= '1';
                ns <= S_WAIT_INIT_NHKEY;
            end if;
        elsif (ps = S_WAIT_INIT_NHKEY) then
            sel_salsa_master <= '1';
            salsa_ready <= '1';
            if (salsa_valid = '1') then
                wr_nhkey <= '1';
                en_ctr   <= '1';
                if (unsigned(ctr) = 6) then
                    ld_ctr <= '1';
                    init_poly_key <= '1';
                    ns <= S_WAIT_INIT_PKEY;
                end if;
            end if;
        elsif (ps = S_WAIT_INIT_PKEY) then
            sel_salsa_master <= '1';
            en_ctr <= '1';
            if (ctr(0) = '1') then
                salsa_ready <= '1';
            end if;
            if (ctr(1 downto 0) = "11") then
                ld_ctr <= '1';
                key_ready <= '1';
                ns <= S_WAIT;
            end if;
        elsif (ps = S_LD_LEN) then
            bdi_ready <= '1';
            if (bdi_valid = '1') then
                en_len    <= '1';
                ns <= S_LD_NPUB;
            end if;
        elsif (ps = S_LD_NPUB) then
            bdi_ready <= '1';
            if (bdi_valid = '1') then
                en_npub <= '1';
                st_hash <= '1';
                ns      <= S_WAIT_HASH;
            end if;
        elsif (ps = S_WAIT_HASH) then
            sel_bus_master <= '1';
            if (hash_done = '1') then
                st_salsa <= '1';
                set_second <= '1';
                if (is_decrypt_r = '1') then
                    if (is_second_r = '0') then
                        st_hash    <= '1';
                    else
                        ns <= S_WAIT_SALSA_TAG;
                    end if;
                else
                    if (is_second_r = '1') then
                        st_output <= '1';
                        ns <= S_WAIT_SALSA_OUT;
                    else
                        ns <= S_WAIT_SALSA_TAG;
                    end if;
                end if;
            end if;
        elsif (ps = S_WAIT_SALSA_TAG) then
            if (salsa_valid = '1') then
                if (is_decrypt_r = '0') then
                    --! Enter this stage only in the end of first pass
                    sel_tag <= '1';
                    en_tag  <= '1';
                    st_hash <= '1';
                    ns      <= S_WAIT_HASH;
                else
                    msg_auth_done <= '1';
                    ns      <= S_WAIT;
                end if;
            end if;
        elsif (ps = S_WAIT_SALSA_OUT) then
            sel_fifo       <= '1';
            sel_bus_master <= '1';
            if (output_done = '1') then
                ns <= S_WR_TAG;
            end if;
        elsif (ps = S_WR_TAG) then
            sel_tag <= '1';
            bdo_valid <= '1';
            if (bdo_ready = '1') then
                ns <= S_WAIT;
            end if;
        end if;
    end process;
end architecture structure;