-------------------------------------------------------------------------------
--! @file       Mult32_PPL.vhd
--! @brief      Unsigned 32-bit pipelind multiplier.
--!
--!             Pipelined Radix-4 unsigned multiplication
--!
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity Mult32_PPL is
    generic (
        G_W             : integer := 32
    );
    port    (
        clk             : in  std_logic;
        --! Input
        aa              : in  std_logic_vector(G_W              -1 downto 0);
        bb              : in  std_logic_vector(G_W              -1 downto 0);
        --! Output
        cc              : out std_logic_vector(G_W*2            -1 downto 0);
        --! Control
        en              : in  std_logic
    );
end entity Mult32_PPL;

architecture structure of Mult32_PPL is
    signal aa3      : unsigned(G_W+2            -1 downto 0);
    signal aa3_r    : unsigned(G_W+2            -1 downto 0);
    signal aa_r     : unsigned(G_W              -1 downto 0);
    signal bb_r     : unsigned(G_W              -1 downto 0);

    type t_pp is array (0 to G_W/2-1) of unsigned(G_W+2     -1 downto 0);
    signal pp       : t_pp;
    signal pp_r     : t_pp;

    type t_st2 is array (0 to G_W/8-1) of unsigned(G_W+8    -1 downto 0);
    signal st2      : t_st2;
    signal st2_r    : t_st2;

    signal st3      : unsigned(G_W*2            -1 downto 0);
    signal st3_r    : unsigned(G_W*2            -1 downto 0);
begin
    process(clk)
    begin
        if rising_edge(clk) then
            if (en = '1') then
                aa3_r <= aa3;
                aa_r <= unsigned(aa);
                bb_r <= unsigned(bb);

                pp_r  <= pp;
                st2_r <= st2;
                st3_r <= st3;
            end if;
        end if;
    end process;

    --! ======================================================================
    --! Stage 0
    --! Pre-computation of 3a
    aa3 <= unsigned('0' & aa & '0') +  unsigned(aa);

    --! ======================================================================
    --! Stage 1
    --! Partial production creation

    gpp: for i in 0 to G_W/2-1 generate
        signal sel : unsigned(1 downto 0);
    begin
        sel <= bb_r(i*2+1 downto i*2);
        with sel select
        pp(i) <= (others => '0')  when "00",
                 "00" & aa_r      when "01",
                 '0' & aa_r & '0' when "10",
                 aa3_r            when others;
    end generate;
            
    --! ======================================================================
    --! Stage 2
    --! 4 - 4:1 reduction
    greduce: for i in 0 to (G_W/2)/4-1 generate
        signal tmp1 : unsigned(G_W+4         -1 downto 0);
        signal tmp2 : unsigned(G_W+4         -1 downto 0);
    begin
        tmp1 <= (pp_r(i*4+1) + pp_r(i*4+0)(G_W+2-1 downto 2)) 
                & pp_r(i*4+0)(1 downto 0);
        tmp2 <= (pp_r(i*4+3) + pp_r(i*4+2)(G_W+2-1 downto 2)) 
                & pp_r(i*4+2)(1 downto 0);

        st2(i) <= (tmp2 + tmp1(G_W+4-1 downto 4)) & tmp1(3 downto 0);
    end generate;

    --! ======================================================================
    --! Stage 3
    --! 1 - 4:1 reduction

    breduce: block
        signal tmp1 : unsigned(G_W+16         -1 downto 0);
        signal tmp2 : unsigned(G_W+16         -1 downto 0);
    begin
        --!
        tmp1(7        downto 0) <= st2_r(0)(7 downto 0);
        tmp1(G_W+16-1 downto 8) <= st2_r(1) + st2_r(0)(G_W+8-1 downto 8);

        tmp2(7        downto 0) <= st2_r(2)(7 downto 0);
        tmp2(G_W+16-1 downto 8) <= st2_r(3) + st2_r(2)(G_W+8-1 downto 8);

        st3(      15 downto  0) <= tmp1(15 downto 0);
        st3(G_W+32-1 downto 16) <= tmp2 + tmp1(G_W+16-1 downto 16);
    end block;

    --! ======================================================================
    --! Stage 4
    --! output
    cc <= std_logic_vector(st3_r);
end architecture structure;