-------------------------------------------------------------------------------
--! @file       Salsa20.vhd
--! @brief      Salsa20 module
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity Salsa20 is
    generic (
        --! Horizontal Folding Factor (1 or 2)
        G_HF                : integer := 2
    );
    port    (
        clk                 : in  std_logic;
        rst                 : in  std_logic;
        --! Input
        xx                  : in  std_logic_vector(512          -1 downto 0);
        --! Control
        start               : in  std_logic;
        almost_done         : out std_logic;
        done                : out std_logic;
        ready               : out std_logic;
        --! Output
        yy                  : out std_logic_vector(512          -1 downto 0)
    );
end entity Salsa20;

architecture structure of Salsa20 is
    constant CHACHA_ROUNDS          : integer := 12;

    procedure qround_half (
        signal a,b,c,d   : in  std_logic_vector(32       -1 downto 0);
        signal sel_round : in  std_logic;
        signal w,x,y,z   : out std_logic_vector(32       -1 downto 0))
    is
        variable tmp_a, tmp_b, tmp_c, tmp_d : unsigned(32       -1 downto 0);
        variable tmp_a2, tmp_b2             : unsigned(32       -1 downto 0);
        variable tmp_c2, tmp_d2             : unsigned(32       -1 downto 0);
        variable rot_d, rot_b               : unsigned(32       -1 downto 0);
        variable rot_d2, rot_b2             : unsigned(32       -1 downto 0);
    begin
        --! a += b; d ^= a; d <<<= 16;
        tmp_a := unsigned(a) + unsigned(b);
        tmp_d := unsigned(d) xor tmp_a;
        if (sel_round = '0') then
            rot_d := rotate_left(tmp_d, natural(16));
        else
            rot_d := rotate_left(tmp_d, natural(8));
        end if;

        --! c += d; b ^= c; b <<<= 12;
        tmp_c := unsigned(c) + rot_d;
        tmp_b := unsigned(b) xor tmp_c;
        if (sel_round = '0') then
            rot_b := rotate_left(tmp_b, natural(12));
        else
            rot_b := rotate_left(tmp_b, natural(7));
        end if;
        
        w <= std_logic_vector(tmp_a);
        x <= std_logic_vector(rot_b);
        y <= std_logic_vector(tmp_c);
        z <= std_logic_vector(rot_d);
    end qround_half;

    procedure quarterround (
        signal a,b,c,d : in  std_logic_vector(32       -1 downto 0);
        signal w,x,y,z : out std_logic_vector(32       -1 downto 0))
    is
        variable tmp_a, tmp_b, tmp_c, tmp_d : unsigned(32       -1 downto 0);
        variable tmp_a2, tmp_b2             : unsigned(32       -1 downto 0);
        variable tmp_c2, tmp_d2             : unsigned(32       -1 downto 0);
        variable rot_d, rot_b               : unsigned(32       -1 downto 0);
        variable rot_d2, rot_b2             : unsigned(32       -1 downto 0);
    begin
        --! a += b; d ^= a; d <<<= 16;
        tmp_a := unsigned(a) + unsigned(b);
        tmp_d := unsigned(d) xor tmp_a;
        rot_d := rotate_left(tmp_d, natural(16));

        --! c += d; b ^= c; b <<<= 12;
        tmp_c := unsigned(c) + rot_d;
        tmp_b := unsigned(b) xor tmp_c;
        rot_b := rotate_left(tmp_b, natural(12));

        --! a += b; d ^= a; d <<<= 8;
        tmp_a2 := tmp_a + rot_b;
        tmp_d2 := rot_d xor tmp_a2;
        rot_d2 := rotate_left(tmp_d2, natural(8));

        --! c += d; b ^= c; b <<<= 7;
        tmp_c2 := tmp_c + rot_d2;
        tmp_b2 := rot_b xor tmp_c2;
        rot_b2 := rotate_left(tmp_b2, natural(7));

        w <= std_logic_vector(tmp_a2);
        x <= std_logic_vector(rot_b2);
        y <= std_logic_vector(tmp_c2);
        z <= std_logic_vector(rot_d2);
    end quarterround;

    --! Control signals
    --!     Status
    signal round                    : std_logic_vector(4+G_HF/2    -1 downto 0);
    --!     from controller
    signal almost_done_s            : std_logic;
    signal almost_done_r            : std_logic;
    signal en_round                 : std_logic;
    signal clr_round                : std_logic;

    --!     State
    type    state_type is (S_IDLE, S_COMPUTE);
    signal state                    : state_type;
    signal nstate                   : state_type;

    --! Data signals
    --!     signals
    type salsa_type is array (0 to 15) of std_logic_vector(32       -1 downto 0);
    signal to_state1                : salsa_type;
    signal to_state2                : salsa_type;
    signal ss                       : salsa_type;
    signal qr                       : salsa_type;
    

    type permute_type is array ( 0 to 15 ) of integer;
	constant PERMUTE : permute_type := ( 0,1,2,3,5,6,7,4,10,11,8,9,15,12,13,14 );
begin
    --! ==============
    --! Datapath
    --! ==============
    --! Registers
    pReg: process(clk)
    begin
        if rising_edge(clk) then
            --! State register
            if (start = '1') then
                ss <= to_state1;
            elsif (en_round = '1') then
                ss <= to_state2;
            end if;
        end if;
    end process;

    gMap: for i in 0 to 15 generate
        to_state1(i) <= xx(512-i*32-1 downto 512-(i+1)*32);
        yy(512-i*32-1 downto 512-(i+1)*32) <= ss(i);
    end generate;

    fh1: if (G_HF/2 = 0) generate
        quarterround(ss( 0),ss( 4),ss( 8),ss(12),qr( 0),qr( 4),qr( 8),qr(12));
        quarterround(ss( 1),ss( 5),ss( 9),ss(13),qr( 1),qr( 5),qr( 9),qr(13));
        quarterround(ss( 2),ss( 6),ss(10),ss(14),qr( 2),qr( 6),qr(10),qr(14));
        quarterround(ss( 3),ss( 7),ss(11),ss(15),qr( 3),qr( 7),qr(11),qr(15));
        to_state2 <= qr;
    end generate;
    fh2: if (G_HF/2 = 1) generate
        signal fw_perm : salsa_type;
        signal bw_perm : salsa_type;
    begin
        qround_half(ss( 0),ss( 4),ss( 8),ss(12),round(0),
                    qr( 0),qr( 4),qr( 8),qr(12));
        qround_half(ss( 1),ss( 5),ss( 9),ss(13),round(0),
                    qr( 1),qr( 5),qr( 9),qr(13));
        qround_half(ss( 2),ss( 6),ss(10),ss(14),round(0),
                    qr( 2),qr( 6),qr(10),qr(14));
        qround_half(ss( 3),ss( 7),ss(11),ss(15),round(0),
                    qr( 3),qr( 7),qr(11),qr(15));
        gPermute: for i in 0 to 15 generate
            fw_perm(i)          <= qr(PERMUTE(i));
            bw_perm(PERMUTE(i)) <= qr(i);
        end generate;
        with round(1 downto 0) select
        to_state2 <=  fw_perm when "01",
                      bw_perm when "11",
                      qr      when others;
    end generate;
    

    --! ==============
    --! Controller
    --! ==============
    pCtrlReg: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then
                state <= S_IDLE;
            else
                state <= nstate;
            end if;

            if (clr_round = '1' or start = '1') then
                round <= (others => '0');
            elsif (en_round = '1') then
                round <= std_logic_vector(unsigned(round) + natural(1));
            end if;
            
            almost_done_r <= almost_done_s;
            done          <= almost_done_r;

            if (rst = '1') then
                ready <= '1';
            elsif (start = '1') then
                ready <= '0';
            elsif (almost_done_r = '1') then
                ready <= '1';
            end if;
        end if;
    end process;
    almost_done <= almost_done_r;

    pCtrlComb: process(state, start, round, almost_done_r)
    begin
        --! internal
        nstate          <= state;
        clr_round       <= '0';
        en_round        <= '0';
        almost_done_s   <= '0';

        case state is
            when S_IDLE =>
                clr_round <= '1';
                if (start = '1') then
                    nstate <= S_COMPUTE;
                end if;

            when S_COMPUTE =>
                en_round <= '1';
                if (almost_done_r = '1') then
                    nstate    <= S_IDLE;
                end if;
                if (unsigned(round) = natural(CHACHA_ROUNDS*G_HF-2)) then
                    almost_done_s <= '1';
                end if;
        end case;
    end process;
end architecture structure;
